@@ -2,6 +2,7 @@
 
 #include "window-item-rec.h"
 
+char *name;
 char *topic;
 char *topic_by;
 time_t topic_time;
@@ -22,4 +23,10 @@ unsigned int synced:1; /* Channel synced - all queries done */
 unsigned int joined:1; /* Have we even received JOIN event for this channel? */
 unsigned int left:1; /* You just left the channel */
 unsigned int kicked:1; /* You just got kicked */
+unsigned int session_rejoin:1; /* This channel was joined with /UPGRADE */
 unsigned int destroying:1;
+
+/* Return the information needed to call SERVER_REC->channels_join() for
+   this channel. Usually just the channel name, but may contain also the
+   channel key. */
+char *(*get_join_data)(CHANNEL_REC *channel);
