@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -32,10 +32,8 @@
 #include "formats.h"
 #include "themes.h"
 #include "translation.h"
-#ifdef HAVE_GLIB2
 #include "recode.h"
 #include "utf8.h"
-#endif
 
 static const char *format_backs = "04261537";
 static const char *format_fores = "kbgcrmyw";
@@ -299,7 +297,7 @@ void format_create_dest_tag(TEXT_DEST_REC *dest, void *server,
 	dest->window = window != NULL ? window :
 		window_find_closest(server, target, level);
 }
-#ifdef HAVE_GLIB2
+
 static int advance (char const **str, gboolean utf8)
 {
 	if (utf8) {
@@ -308,28 +306,24 @@ static int advance (char const **str, gboolean utf8)
 		c = g_utf8_get_char(*str);
 		*str = g_utf8_next_char(*str);
 
-		return utf8_width(c);
+		return unichar_isprint(c) ? mk_wcwidth(c) : 1;
 	} else {
 		*str += 1;
 
 		return 1;
 	}
 }
-#endif
+
 /* Return length of text part in string (ie. without % codes) */
 int format_get_length(const char *str)
 {
 	GString *tmp;
 	int len;
-#ifdef HAVE_GLIB2
 	gboolean utf8;
-#endif
 
 	g_return_val_if_fail(str != NULL, 0);
 
-#ifdef HAVE_GLIB2
 	utf8 = is_utf8() && g_utf8_validate(str, -1, NULL);
-#endif
 
 	tmp = g_string_new(NULL);
 	len = 0;
@@ -346,12 +340,8 @@ int format_get_length(const char *str)
 			if (*str != '%')
 				len++;
 		}
-#ifdef HAVE_GLIB2
+
 		len += advance(&str, utf8);
-#else
-	  len++;
-		str++;
-#endif
 	}
 
 	g_string_free(tmp, TRUE);
@@ -365,16 +355,13 @@ int format_real_length(const char *str, int len)
 {
 	GString *tmp;
 	const char *start;
-#ifdef HAVE_GLIB2
 	const char *oldstr;
 	gboolean utf8;
-#endif
+
 	g_return_val_if_fail(str != NULL, 0);
 	g_return_val_if_fail(len >= 0, 0);
 
-#ifdef HAVE_GLIB2
 	utf8 = is_utf8() && g_utf8_validate(str, -1, NULL);
-#endif
 
 	start = str;
 	tmp = g_string_new(NULL);
@@ -394,15 +381,10 @@ int format_real_length(const char *str, int len)
 			}
 		}
 
-#ifdef HAVE_GLIB2
 		oldstr = str;
 		len -= advance(&str, utf8);
 		if (len < 0)
 			str = oldstr;
-#else
-	  len--;
-		str++;
-#endif
 	}
 
 	g_string_free(tmp, TRUE);
