@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -571,11 +571,8 @@ int key_pressed(KEYBOARD_REC *keyboard, const char *key)
                 g_strconcat(keyboard->key_state, "-", key, NULL);
 	g_free_and_null(keyboard->key_state);
 
-#if GLIB_MAJOR_VERSION == 2
-#  define GSearchFunc GCompareFunc
-#endif
 	rec = g_tree_search(key_states,
-			    (GSearchFunc) key_states_search,
+			    (GCompareFunc) key_states_search,
 			    combo);
 	if (rec == NULL) {
 		/* unknown key combo, eat the invalid key
