@@ -18,6 +18,10 @@ int utf16_char_to_utf8(unichar c, char *outbuf);
    Make sure out is at least 6 x length of str. */
 void utf16_to_utf8(const unichar *str, char *out);
 
+/* unichar -> UTF-8 string with position transformed. The NUL is copied as well.
+   Make sure out is at least 6 x length of str. */
+void utf16_to_utf8_with_pos(const unichar *str, int spos, char *out, int *opos);
+
 /* XXX I didn't check the encoding range of big5+. This is standard big5. */
 #define is_big5_los(lo) (0x40 <= (lo) && (lo) <= 0x7E) /* standard */
 #define is_big5_lox(lo) (0x80 <= (lo) && (lo) <= 0xFE) /* extended */
@@ -26,6 +30,8 @@ void utf16_to_utf8(const unichar *str, char *out);
 #define is_big5(hi,lo) (is_big5_hi(hi) && is_big5_lo(lo))
 
 /* Returns width for character (0-2). */
-int utf8_width(unichar c);
+int mk_wcwidth(unichar c);
+
+#define unichar_isprint(c) (((c) & ~0x80) >= 32)
 
 #endif
