@@ -130,7 +130,8 @@ enum {
 	IRCTXT_SILENCED,
 	IRCTXT_UNSILENCED,
 	IRCTXT_SILENCE_LINE,
-        IRCTXT_ASK_OPER_PASS
+        IRCTXT_ASK_OPER_PASS,
+	IRCTXT_ACCEPT_LIST
 };
 
 extern FORMAT_REC fecommon_irc_formats[];
