@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -31,22 +31,10 @@
 #include "gui-windows.h"
 #include "gui-printtext.h"
 
+#define DEFAULT_LASTLOG_BEFORE 3
+#define DEFAULT_LASTLOG_AFTER 3
 #define MAX_LINES_WITHOUT_FORCE 1000
 
-static void window_lastlog_clear(WINDOW_REC *window)
-{
-	GList *tmp, *next;
-
-	for (tmp = WINDOW_GUI(window)->lines; tmp != NULL; tmp = next) {
-		LINE_REC *line = tmp->data;
-
-                next = tmp->next;
-                if (line->level & MSGLEVEL_LASTLOG)
-			gui_window_line_remove(window, line, FALSE);
-	}
-        gui_window_redraw(window);
-}
-
 /* Only unknown keys in `optlist' should be levels.
    Returns -1 if unknown option was given. */
 int cmd_options_get_level(const char *cmd, GHashTable *optlist)
@@ -87,20 +75,21 @@ int cmd_options_get_level(const char *cmd, GHashTable *optlist)
 }
 
 static void show_lastlog(const char *searchtext, GHashTable *optlist,
-			 int start, int count, int fhandle)
+			 int start, int count, FILE *fhandle)
 {
-        WINDOW_REC *window;
-	GList *startline, *list, *tmp;
+	WINDOW_REC *window;
+        LINE_REC *startline;
+	GList *list, *tmp;
 	GString *line;
         char *str;
-	int level, len;
+	int level, before, after, len;
 
         level = cmd_options_get_level("lastlog", optlist);
 	if (level == -1) return; /* error in options */
         if (level == 0) level = MSGLEVEL_ALL;
 
 	if (g_hash_table_lookup(optlist, "clear") != NULL) {
-		window_lastlog_clear(active_win);
+		textbuffer_view_remove_lines_by_level(WINDOW_GUI(active_win)->view, MSGLEVEL_LASTLOG);
 		if (*searchtext == '\0')
                         return;
 	}
@@ -120,16 +109,32 @@ static void show_lastlog(const char *searchtext, GHashTable *optlist,
 	}
 
 	if (g_hash_table_lookup(optlist, "new") != NULL)
-		startline = WINDOW_GUI(window)->lastlog_last_check;
+		startline = textbuffer_view_get_bookmark(WINDOW_GUI(window)->view, "lastlog_last_check");
 	else if (g_hash_table_lookup(optlist, "away") != NULL)
-		startline = WINDOW_GUI(window)->lastlog_last_away;
+		startline = textbuffer_view_get_bookmark(WINDOW_GUI(window)->view, "lastlog_last_away");
 	else
 		startline = NULL;
-	if (startline == NULL) startline = WINDOW_GUI(window)->lines;
 
-	list = gui_window_find_text(window, startline,
+	if (startline == NULL)
+                startline = textbuffer_view_get_lines(WINDOW_GUI(window)->view);
+
+	str = g_hash_table_lookup(optlist, "#");
+	if (str != NULL) {
+		before = after = atoi(str);
+	} else {
+		str = g_hash_table_lookup(optlist, "before");
+		before = str == NULL ? 0 : *str != '\0' ?
+			atoi(str) : DEFAULT_LASTLOG_BEFORE;
+
+		str = g_hash_table_lookup(optlist, "after");
+		if (str == NULL) str = g_hash_table_lookup(optlist, "a");
+		after = str == NULL ? 0 : *str != '\0' ?
+			atoi(str) : DEFAULT_LASTLOG_AFTER;
+	}
+
+	list = textbuffer_find_text(WINDOW_GUI(window)->view->buffer, startline,
 				    level, MSGLEVEL_LASTLOG,
-				    searchtext,
+				    searchtext, before, after,
 				    g_hash_table_lookup(optlist, "regexp") != NULL,
 				    g_hash_table_lookup(optlist, "word") != NULL,
 				    g_hash_table_lookup(optlist, "case") != NULL);
@@ -138,34 +143,54 @@ static void show_lastlog(const char *searchtext, GHashTable *optlist,
 	if (count <= 0)
 		tmp = list;
 	else {
-		int pos = len-count;
-
+		int pos = len-count-start;
 		if (pos < 0) pos = 0;
-		pos += start;
 
 		tmp = pos > len ? NULL : g_list_nth(list, pos);
 		len = g_list_length(tmp);
 	}
 
-	if (len > MAX_LINES_WITHOUT_FORCE && fhandle == -1 &&
-	    g_hash_table_lookup(optlist, "force") == NULL) {
+	if (g_hash_table_lookup(optlist, "count") != NULL) {
 		printformat_window(active_win, MSGLEVEL_CLIENTNOTICE,
+				   TXT_LASTLOG_COUNT, len);
+		g_list_free(list);
+		return;
+	}
+
+	if (len > MAX_LINES_WITHOUT_FORCE && fhandle == NULL &&
+	    g_hash_table_lookup(optlist, "force") == NULL) {
+		printformat_window(active_win,
+				   MSGLEVEL_CLIENTNOTICE|MSGLEVEL_LASTLOG,
 				   TXT_LASTLOG_TOO_LONG, len);
 		g_list_free(list);
 		return;
 	}
 
-	if (fhandle == -1 && g_hash_table_lookup(optlist, "-") == NULL)
+	if (fhandle == NULL && g_hash_table_lookup(optlist, "-") == NULL)
 		printformat(NULL, NULL, MSGLEVEL_LASTLOG, TXT_LASTLOG_START);
 
 	line = g_string_new(NULL);
         while (tmp != NULL && (count < 0 || count > 0)) {
 		LINE_REC *rec = tmp->data;
 
+		if (rec == NULL) {
+			if (tmp->next == NULL)
+                                break;
+			if (fhandle != NULL) {
+				fwrite("--\n", 3, 1, fhandle);
+			} else {
+				printformat_window(active_win,
+						   MSGLEVEL_LASTLOG,
+						   TXT_LASTLOG_SEPARATOR);
+			}
+                        tmp = tmp->next;
+			continue;
+		}
+
                 /* get the line text */
-		gui_window_line2text(rec, fhandle == -1, line);
+		textbuffer_line2text(rec, fhandle == NULL, line);
 		if (!settings_get_bool("timestamps")) {
-			struct tm *tm = localtime(&rec->time);
+			struct tm *tm = localtime(&rec->info.time);
                         char timestamp[10];
 
 			g_snprintf(timestamp, sizeof(timestamp),
@@ -175,9 +200,9 @@ static void show_lastlog(const char *searchtext, GHashTable *optlist,
 		}
 
                 /* write to file/window */
-		if (fhandle != -1) {
-			write(fhandle, line->str, line->len);
-			write(fhandle, "\n", 1);
+		if (fhandle != NULL) {
+			fwrite(line->str, line->len, 1, fhandle);
+			fputc('\n', fhandle);
 		} else {
 			printtext_window(active_win, MSGLEVEL_LASTLOG,
 					 "%s", line->str);
@@ -188,34 +213,38 @@ static void show_lastlog(const char *searchtext, GHashTable *optlist,
 	}
         g_string_free(line, TRUE);
 
-	if (fhandle == -1 && g_hash_table_lookup(optlist, "-") == NULL)
+	if (fhandle == NULL && g_hash_table_lookup(optlist, "-") == NULL)
 		printformat(NULL, NULL, MSGLEVEL_LASTLOG, TXT_LASTLOG_END);
 
-	WINDOW_GUI(window)->lastlog_last_check =
-		g_list_last(WINDOW_GUI(window)->bottom_startline);
+	textbuffer_view_set_bookmark_bottom(WINDOW_GUI(window)->view,
+					    "lastlog_last_check");
 
+        textbuffer_line_unref_list(WINDOW_GUI(window)->view->buffer, list);
 	g_list_free(list);
 }
 
-/* SYNTAX: LASTLOG [-] [-file <filename>] [-clear] [-<level> -<level...>]
-		   [-new | -away] [-regexp | -word] [-case]
-		   [-window <ref#|name>] [<pattern>] [<count> [<start>]] */
+/* SYNTAX: LASTLOG [-] [-file <filename>] [-window <ref#|name>] [-new | -away]
+		   [-<level> -<level...>] [-clear] [-count] [-case]
+		   [-regexp | -word] [-before [<#>]] [-after [<#>]]
+		   [-<# before+after>] [<pattern>] [<count> [<start>]] */
 static void cmd_lastlog(const char *data)
 {
 	GHashTable *optlist;
 	char *text, *countstr, *start, *fname;
 	void *free_arg;
-        int count, fhandle;
+        int count, fd;
+	FILE *fhandle;
 
 	g_return_if_fail(data != NULL);
 
-	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_OPTIONS | PARAM_FLAG_UNKNOWN_OPTIONS,
-			    "lastlog", &optlist, &text, &countstr, &start))
+	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_OPTIONS |
+			    PARAM_FLAG_UNKNOWN_OPTIONS, "lastlog", &optlist,
+			    &text, &countstr, &start))
 		return;
 
-	if (*start == '\0' && is_numeric(text, 0)) {
-		if (is_numeric(countstr, 0))
-			start = countstr;
+	if (*start == '\0' && is_numeric(text, 0) && *text != '0' &&
+	    (*countstr == '\0' || is_numeric(countstr, 0))) {
+		start = countstr;
 		countstr = text;
 		text = "";
 	}
@@ -223,22 +252,31 @@ static void cmd_lastlog(const char *data)
 	if (count == 0) count = -1;
 
 	/* target where to print it */
-        fhandle = -1;
+        fhandle = NULL;
 	fname = g_hash_table_lookup(optlist, "file");
 	if (fname != NULL) {
                 fname = convert_home(fname);
-		fhandle = open(fname, O_WRONLY | O_APPEND | O_CREAT,
-			       octal2dec(settings_get_int("log_create_mode")));
+		fd = open(fname, O_WRONLY | O_APPEND | O_CREAT,
+			  octal2dec(settings_get_int("log_create_mode")));
+		if (fd != -1) {
+			fhandle = fdopen(fd, "a");
+			if (fhandle == NULL)
+				close(fd);
+		}
                 g_free(fname);
 	}
 
-	if (fname != NULL && fhandle == -1) {
+	if (fname != NULL && fhandle == NULL) {
 		printtext(NULL, NULL, MSGLEVEL_CLIENTERROR,
-			  "%s", g_strerror(errno));
+			  "Could not open lastlog: %s", g_strerror(errno));
 	} else {
 		show_lastlog(text, optlist, atoi(start), count, fhandle);
-		if (fhandle != -1)
-			close(fhandle);
+		if (fhandle != NULL) {
+			if (ferror(fhandle))
+				printtext(NULL, NULL, MSGLEVEL_CLIENTERROR,
+				  	  "Could not write lastlog: %s", g_strerror(errno));
+			fclose(fhandle);
+		}
 	}
 
 	cmd_params_free(free_arg);
@@ -248,7 +286,7 @@ void lastlog_init(void)
 {
 	command_bind("lastlog", NULL, (SIGNAL_FUNC) cmd_lastlog);
 
-	command_set_options("lastlog", "!- force clear -file -window new away word regexp case");
+	command_set_options("lastlog", "!- # force clear -file -window new away word regexp case count @a @after @before");
 }
 
 void lastlog_deinit(void)
