@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -23,18 +23,15 @@
 #include "misc.h"
 #include "channels-setup.h"
 #include "settings.h"
+#include "recode.h"
 
 #include "irc-servers.h"
 #include "irc-channels.h"
 
-static void event_cannot_join(IRC_SERVER_REC *server, const char *data)
+static void check_join_failure(IRC_SERVER_REC *server, const char *channel)
 {
 	CHANNEL_REC *chanrec;
-	char *params, *channel;
-
-	g_return_if_fail(data != NULL);
-
-	params = event_get_params(data, 2, NULL, &channel);
+	char *chan2;
 
 	if (channel[0] == '!' && channel[1] == '!')
 		channel++; /* server didn't understand !channels */
@@ -43,15 +40,28 @@ static void event_cannot_join(IRC_SERVER_REC *server, const char *data)
 	if (chanrec == NULL && channel[0] == '!') {
 		/* it probably replied with the full !channel name,
 		   find the channel with the short name.. */
-		channel = g_strdup_printf("!%s", channel+6);
-		chanrec = channel_find(SERVER(server), channel);
-		g_free(channel);
+		chan2 = g_strdup_printf("!%s", channel+6);
+		chanrec = channel_find(SERVER(server), chan2);
+		g_free(chan2);
 	}
 
 	if (chanrec != NULL && !chanrec->joined) {
 		chanrec->left = TRUE;
 		channel_destroy(chanrec);
 	}
+}
+
+static void irc_server_event(IRC_SERVER_REC *server, const char *line)
+{
+	char *params, *numeric, *channel;
+
+	/* We'll be checking "4xx <your nick> <channel>" for channels
+	   which we haven't joined yet. 4xx are error codes and should
+	   indicate that the join failed. */
+	params = event_get_params(line, 3, &numeric, NULL, &channel);
+
+	if (numeric[0] == '4')
+		check_join_failure(server, channel);
 
 	g_free(params);
 }
@@ -65,21 +75,21 @@ static void event_no_such_channel(IRC_SERVER_REC *server, const char *data)
 	params = event_get_params(data, 2, NULL, &channel);
 	chanrec = *channel == '!' && channel[1] != '\0' ?
 		channel_find(SERVER(server), channel) : NULL;
-	g_free(params);
 
 	if (chanrec != NULL) {
                 /* !channel didn't exist, so join failed */
 		setup = channel_setup_find(chanrec->name,
 					   chanrec->server->connrec->chatnet);
 		if (setup != NULL && setup->autojoin) {
-			/* it's autojoin channel though,
-			   so create it */
+			/* it's autojoin channel though, so create it */
 			irc_send_cmdv(server, "JOIN !%s", chanrec->name);
+			g_free(params);
                         return;
 		}
 	}
 
-        event_cannot_join(server, data);
+	check_join_failure(server, channel);
+	g_free(params);
 }
 
 static void event_duplicate_channel(IRC_SERVER_REC *server, const char *data)
@@ -107,35 +117,24 @@ static void event_duplicate_channel(IRC_SERVER_REC *server, const char *data)
 	g_free(params);
 }
 
-static void event_target_unavailable(IRC_SERVER_REC *server, const char *data)
-{
-	char *params, *channel;
-
-	g_return_if_fail(data != NULL);
-
-	params = event_get_params(data, 2, NULL, &channel);
-	if (ischannel(*channel)) {
-		/* channel is unavailable - try to join again a bit later */
-		event_cannot_join(server, data);
-	}
-
-	g_free(params);
-}
-
 static void channel_change_topic(IRC_SERVER_REC *server, const char *channel,
 				 const char *topic, const char *setby,
 				 time_t settime)
 {
 	CHANNEL_REC *chanrec;
-
+	char *recoded = NULL;
+	
 	chanrec = channel_find(SERVER(server), channel);
 	if (chanrec == NULL) return;
-
+	/* the topic may be send out encoded, so we need to 
+	   recode it back or /topic <tab> will not work properly */
+	recoded = recode_in(SERVER(server), topic, channel);
 	if (topic != NULL) {
 		g_free_not_null(chanrec->topic);
-		chanrec->topic = *topic == '\0' ? NULL : g_strdup(topic);
+		chanrec->topic = recoded == NULL ? NULL : g_strdup(recoded);
 	}
-	
+	g_free(recoded);
+
 	g_free_not_null(chanrec->topic_by);
 	chanrec->topic_by = g_strdup(setby);
 	
@@ -253,12 +252,11 @@ static void event_join(IRC_SERVER_REC *server, const char *data, const char *nic
 
 	chanrec = irc_channel_find(server, channel);
 	if (chanrec != NULL && chanrec->joined) {
-		/* already joined this channel - this check was added
-		   here because of broken irssi proxy :) */
-		g_free(shortchan);
-		g_free(params);
-                return;
-	}
+		/* already joined this channel - probably a broken proxy that
+		   forgot to send PART between */
+		chanrec->left = TRUE;
+		channel_destroy(CHANNEL(chanrec));
+		chanrec = NULL;	}
 
 	if (chanrec == NULL) {
 		/* look again, because of the channel name cut issues. */
@@ -361,18 +359,9 @@ void channel_events_init(void)
 {
 	settings_add_bool("misc", "join_auto_chans_on_invite", TRUE);
 
+	signal_add_last("server event", (SIGNAL_FUNC) irc_server_event);
 	signal_add_first("event 403", (SIGNAL_FUNC) event_no_such_channel); /* no such channel */
-	signal_add_first("event 405", (SIGNAL_FUNC) event_cannot_join); /* too many channels */
 	signal_add_first("event 407", (SIGNAL_FUNC) event_duplicate_channel); /* duplicate channel */
-	signal_add_first("event 442", (SIGNAL_FUNC) event_cannot_join); /* not on that channel (dalnet) */
-	signal_add_first("event 477", (SIGNAL_FUNC) event_cannot_join); /* need to be identify to registered nick (dalnet) */
-	signal_add_first("event 471", (SIGNAL_FUNC) event_cannot_join); /* channel is full */
-	signal_add_first("event 473", (SIGNAL_FUNC) event_cannot_join); /* invite only */
-	signal_add_first("event 474", (SIGNAL_FUNC) event_cannot_join); /* banned */
-	signal_add_first("event 475", (SIGNAL_FUNC) event_cannot_join); /* bad channel key */
-	signal_add_first("event 476", (SIGNAL_FUNC) event_cannot_join); /* bad channel mask */
-	signal_add_first("event 379", (SIGNAL_FUNC) event_cannot_join); /* forwarding to another channel */
-	signal_add_first("event 439", (SIGNAL_FUNC) event_cannot_join); /* target change too fast */
 
 	signal_add("event topic", (SIGNAL_FUNC) event_topic);
 	signal_add_first("event join", (SIGNAL_FUNC) event_join);
@@ -381,23 +370,13 @@ void channel_events_init(void)
 	signal_add("event invite", (SIGNAL_FUNC) event_invite);
 	signal_add("event 332", (SIGNAL_FUNC) event_topic_get);
 	signal_add("event 333", (SIGNAL_FUNC) event_topic_info);
-	signal_add_first("event 437", (SIGNAL_FUNC) event_target_unavailable); /* channel/nick unavailable */
 }
 
 void channel_events_deinit(void)
 {
+	signal_remove("server event", (SIGNAL_FUNC) irc_server_event);
 	signal_remove("event 403", (SIGNAL_FUNC) event_no_such_channel); /* no such channel */
-	signal_remove("event 405", (SIGNAL_FUNC) event_cannot_join); /* too many channels */
 	signal_remove("event 407", (SIGNAL_FUNC) event_duplicate_channel); /* duplicate channel */
-	signal_remove("event 442", (SIGNAL_FUNC) event_cannot_join); /* not on that channel (dalnet) */
-	signal_remove("event 477", (SIGNAL_FUNC) event_cannot_join); /* need to be identify to registered nick (dalnet) */
-	signal_remove("event 471", (SIGNAL_FUNC) event_cannot_join); /* channel is full */
-	signal_remove("event 473", (SIGNAL_FUNC) event_cannot_join); /* invite only */
-	signal_remove("event 474", (SIGNAL_FUNC) event_cannot_join); /* banned */
-	signal_remove("event 475", (SIGNAL_FUNC) event_cannot_join); /* bad channel key */
-	signal_remove("event 476", (SIGNAL_FUNC) event_cannot_join); /* bad channel mask */
-	signal_remove("event 379", (SIGNAL_FUNC) event_cannot_join); /* forwarding to another channel */
-	signal_remove("event 439", (SIGNAL_FUNC) event_cannot_join); /* target change too fast */
 
 	signal_remove("event topic", (SIGNAL_FUNC) event_topic);
 	signal_remove("event join", (SIGNAL_FUNC) event_join);
@@ -406,5 +385,4 @@ void channel_events_deinit(void)
 	signal_remove("event invite", (SIGNAL_FUNC) event_invite);
 	signal_remove("event 332", (SIGNAL_FUNC) event_topic_get);
 	signal_remove("event 333", (SIGNAL_FUNC) event_topic_info);
-	signal_remove("event 437", (SIGNAL_FUNC) event_target_unavailable); /* channel/nick unavailable */
 }
