@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -577,7 +577,10 @@ static void cmd_quote(const char *data, IRC_SERVER_REC *server)
 	if (server == NULL || server->connect_time == 0)
 		cmd_return_error(CMDERR_NOT_CONNECTED);
 
-	irc_send_cmd(server, data);
+	if (!server->connected)
+		irc_send_cmd_now(server, data);
+	else
+		irc_send_cmd(server, data);
 }
 
 /* SYNTAX: RAWQUOTE <data> */
@@ -799,6 +802,7 @@ static void cmd_knockout(const char *data, IRC_SERVER_REC *server,
         char **nicklist, *spacenicks, *banmasks;
 	void *free_arg;
 	int timeleft;
+	GSList *ptr;
 
         CMD_IRC_SERVER(server);
 
@@ -853,13 +857,20 @@ static void cmd_knockout(const char *data, IRC_SERVER_REC *server,
 	if (*banmasks == '\0')
 		g_free(banmasks);
 	else {
-		/* create knockout record */
-		rec = g_new(KNOCKOUT_REC, 1);
+		/* check if we already have this knockout */
+		for (ptr = server->knockoutlist; ptr != NULL; ptr = ptr->next) {
+			rec = ptr->data;
+			if (channel == rec->channel &&
+					!strcmp(rec->ban, banmasks))
+				break;
+		}
+		if (ptr == NULL) {
+			rec = g_new(KNOCKOUT_REC, 1);
+			rec->channel = channel;
+			rec->ban = banmasks;
+			server->knockoutlist = g_slist_append(server->knockoutlist, rec);
+		}
 		rec->unban_time = time(NULL)+timeleft/1000;
-		rec->channel = channel;
-		rec->ban = banmasks;
-
-		server->knockoutlist = g_slist_append(server->knockoutlist, rec);
 	}
 
 	cmd_params_free(free_arg);
@@ -929,6 +940,17 @@ static void cmd_oper(const char *data, IRC_SERVER_REC *server)
 	cmd_params_free(free_arg);
 }
 
+/* SYNTAX: ACCEPT [[-]nick,...] */
+static void cmd_accept(const char *data, IRC_SERVER_REC *server)
+{
+        CMD_IRC_SERVER(server);
+
+	if (*data == '\0') 
+		irc_send_cmd(server, "ACCEPT *");
+	else
+		irc_send_cmdv(server, "ACCEPT %s", data);
+}
+
 /* SYNTAX: UNSILENCE <nick!user@host> */
 static void cmd_unsilence(const char *data, IRC_SERVER_REC *server)
 {
@@ -1004,6 +1026,7 @@ void irc_commands_init(void)
 	command_bind_irc("away", NULL, (SIGNAL_FUNC) cmd_away);
 	/* SYNTAX: ISON <nicks> */
 	command_bind_irc("ison", NULL, (SIGNAL_FUNC) command_1self);
+	command_bind_irc("accept", NULL, (SIGNAL_FUNC) cmd_accept);
 	/* SYNTAX: ADMIN [<server>|<nickname>] */
 	command_bind_irc("admin", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: INFO [<server>] */
@@ -1018,7 +1041,7 @@ void irc_commands_init(void)
 	command_bind_irc("map", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: MOTD [<server>|<nick>] */
 	command_bind_irc("motd", NULL, (SIGNAL_FUNC) command_self);
-	/* SYNTAX: REHASH */
+	/* SYNTAX: REHASH [<option>] */
 	command_bind_irc("rehash", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: STATS <type> [<server>] */
 	command_bind_irc("stats", NULL, (SIGNAL_FUNC) command_self);
@@ -1099,6 +1122,7 @@ void irc_commands_deinit(void)
 	command_unbind("kill", (SIGNAL_FUNC) command_2self);
 	command_unbind("away", (SIGNAL_FUNC) cmd_away);
 	command_unbind("ison", (SIGNAL_FUNC) command_1self);
+	command_unbind("accept", (SIGNAL_FUNC) cmd_accept);
 	command_unbind("admin", (SIGNAL_FUNC) command_self);
 	command_unbind("info", (SIGNAL_FUNC) command_self);
     command_unbind("knock", (SIGNAL_FUNC) command_self);
