@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -251,8 +251,6 @@ void modes_type_prefix(IRC_CHANNEL_REC *channel, const char *setby,
 		       char type, char mode, char *arg, GString *newmode)
 {
 	int umode = (unsigned char) mode;
-	nick_mode_change(channel, arg, channel->server->modes[umode].prefix,
-			 type, setby);
 
 	if (g_strcasecmp(channel->server->nick, arg) == 0) {
 		/* see if we need to update channel->chanop */
@@ -274,6 +272,9 @@ void modes_type_prefix(IRC_CHANNEL_REC *channel, const char *setby,
 				channel->chanop = type == '+';
 		}
 	}
+
+	nick_mode_change(channel, arg, channel->server->modes[umode].prefix,
+			 type, setby);
 }
 
 int channel_mode_is_set(IRC_CHANNEL_REC *channel, char mode)
