@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -408,7 +408,7 @@ static int dcc_send_one_file(int queue, const char *target, const char *fname,
 	fname = g_basename(fname);
 
 	/* Replace all the spaces with underscore so that lesser
-	   intellgent clients can communicate.. */
+	   intelligent clients can communicate.. */
 	if (!settings_get_bool("dcc_send_replace_space_with_underscore"))
 		str = NULL;
 	else {
