@@ -13,9 +13,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "module.h"
@@ -36,8 +36,7 @@ typedef struct {
 	char *target;
 	int level;
 
-	int msgcount;
-	time_t first;
+	GSList *msgtimes;
 } FLOOD_ITEM_REC;
 
 typedef struct {
@@ -51,7 +50,7 @@ static int flood_max_msgs, flood_timecheck;
 static int flood_hash_check_remove(const char *key, FLOOD_REC *flood,
 				   time_t *now)
 {
-	GSList *tmp, *next;
+	GSList *tmp, *next, *times, *tnext;
 
 	g_return_val_if_fail(key != NULL, FALSE);
 	g_return_val_if_fail(flood != NULL, FALSE);
@@ -60,7 +59,18 @@ static int flood_hash_check_remove(const char *key, FLOOD_REC *flood,
 		FLOOD_ITEM_REC *rec = tmp->data;
 
 		next = tmp->next;
-		if (*now-rec->first >= flood_timecheck) {
+		/* remove old time entries for current rec item */
+		for (times = rec->msgtimes; times != NULL; times = tnext) {
+                        time_t *data = times->data;
+			tnext = times->next;
+
+			if (*now-*((time_t *) times->data) >= flood_timecheck) {
+				rec->msgtimes = g_slist_remove(rec->msgtimes, data);
+				g_free(data);
+			}
+		}
+		/* if no more time entries remove rec item */
+		if (rec->msgtimes == NULL) {
 			flood->items = g_slist_remove(flood->items, rec);
 			g_free(rec->target);
 			g_free(rec);
@@ -79,7 +89,7 @@ static int flood_timeout(void)
 {
 	MODULE_SERVER_REC *mserver;
 	GSList *tmp;
-        time_t now;
+	time_t now;
 
 	/* remove the old people from flood lists */
 	now = time(NULL);
@@ -119,10 +129,12 @@ static void flood_hash_destroy(const char *key, FLOOD_REC *flood)
 	while (flood->items != NULL) {
 		FLOOD_ITEM_REC *rec = flood->items->data;
 
+		flood->items = g_slist_remove(flood->items, rec);
+
+		g_slist_foreach(rec->msgtimes, (GFunc) g_free, NULL);
+		g_slist_free(rec->msgtimes);
 		g_free(rec->target);
 		g_free(rec);
-
-		flood->items = g_slist_remove(flood->items, rec);
 	}
 
 	g_free(flood->nick);
@@ -148,6 +160,7 @@ static void flood_deinit_server(IRC_SERVER_REC *server)
 		g_hash_table_destroy(mserver->floodlist);
 	}
 	g_free(mserver);
+	MODULE_DATA_UNSET(server);
 }
 
 static FLOOD_ITEM_REC *flood_find(FLOOD_REC *flood, int level,
@@ -173,6 +186,7 @@ static void flood_newmsg(IRC_SERVER_REC *server, int level, const char *nick,
 	MODULE_SERVER_REC *mserver;
 	FLOOD_REC *flood;
 	FLOOD_ITEM_REC *rec;
+	time_t *ttime;
 
 	g_return_if_fail(server != NULL);
 	g_return_if_fail(nick != NULL);
@@ -182,9 +196,11 @@ static void flood_newmsg(IRC_SERVER_REC *server, int level, const char *nick,
 
 	rec = flood == NULL ? NULL : flood_find(flood, level, target);
 	if (rec != NULL) {
-		if (++rec->msgcount > flood_max_msgs) {
+		ttime = g_new(time_t, 1);
+		*ttime = time(NULL);
+		rec->msgtimes = g_slist_append(rec->msgtimes, ttime);
+		if (g_slist_length(rec->msgtimes) > flood_max_msgs) {
 			/* flooding! */
-                        rec->first = 0;
 			signal_emit("flood", 5, server, nick, host,
 				    GINT_TO_POINTER(rec->level), target);
 		}
@@ -199,8 +215,10 @@ static void flood_newmsg(IRC_SERVER_REC *server, int level, const char *nick,
 
 	rec = g_new0(FLOOD_ITEM_REC, 1);
 	rec->level = level;
-	rec->first = time(NULL);
-	rec->msgcount = 1;
+	rec->msgtimes = NULL;
+	ttime = g_new(time_t, 1);
+	*ttime = time(NULL);
+	rec->msgtimes = g_slist_append(rec->msgtimes, ttime);
 	rec->target = g_strdup(target);
 
 	flood->items = g_slist_append(flood->items, rec);
@@ -267,16 +285,21 @@ static void read_settings(void)
 	flood_timecheck = settings_get_int("flood_timecheck");
 	flood_max_msgs = settings_get_int("flood_max_msgs");
 
-	if (flood_tag != -1) {
+	if (flood_timecheck > 0 && flood_max_msgs > 0) {
+		if (flood_tag == -1) {
+			flood_tag = g_timeout_add(500, (GSourceFunc) flood_timeout, NULL);
+
+			signal_add("event privmsg", (SIGNAL_FUNC) flood_privmsg);
+			signal_add("event notice", (SIGNAL_FUNC) flood_notice);
+			signal_add("ctcp msg", (SIGNAL_FUNC) flood_ctcp);
+		}
+	} else if (flood_tag != -1) {
 		g_source_remove(flood_tag);
 		flood_tag = -1;
-	}
 
-	if (flood_timecheck > 0 && flood_max_msgs > 0) {
-		flood_tag = g_timeout_add(500, (GSourceFunc) flood_timeout, NULL);
-		signal_add("event privmsg", (SIGNAL_FUNC) flood_privmsg);
-		signal_add("event notice", (SIGNAL_FUNC) flood_notice);
-		signal_add("ctcp msg", (SIGNAL_FUNC) flood_ctcp);
+		signal_remove("event privmsg", (SIGNAL_FUNC) flood_privmsg);
+		signal_remove("event notice", (SIGNAL_FUNC) flood_notice);
+		signal_remove("ctcp msg", (SIGNAL_FUNC) flood_ctcp);
 	}
 }
 
@@ -292,6 +315,8 @@ void irc_flood_init(void)
 	signal_add("server disconnected", (SIGNAL_FUNC) flood_deinit_server);
 
 	autoignore_init();
+	settings_check();
+	module_register("flood", "irc");
 }
 
 void irc_flood_deinit(void)
