@@ -21,15 +21,15 @@ typedef struct {
 	LINEBUF_REC *buffer;
 
 	char *nick, *host;
-	GIOChannel *handle;
-	int tag;
-
+	NET_SENDBUF_REC *handle;
+	int recv_tag;
 	char *proxy_address;
 	LISTEN_REC *listen;
 	IRC_SERVER_REC *server;
 	unsigned int pass_sent:1;
 	unsigned int user_sent:1;
 	unsigned int connected:1;
+	unsigned int want_ctcp:1;
 } CLIENT_REC;
 
 extern GSList *proxy_listens;
